#include <assert.h>
#include "avistream.h"

TAVI_Stream *p_AVI_Load(const char *p_s_avi_file, int b_rgba)
{
	TAVI_Stream t_tmp;
	AVISTREAMINFO t_avi_stream_info;

	t_tmp.p_avi_frame = NULL;
	t_tmp.p_avi_stream = NULL;

	for(;;) {
		HRESULT hr;

		AVIFileInit();

		if(FAILED(hr = AVIStreamOpenFromFile(&t_tmp.p_avi_stream, p_s_avi_file, streamtypeVIDEO, 0, OF_READ, 0)))
			break;
		if(FAILED(hr = AVIStreamInfo(t_tmp.p_avi_stream, &t_avi_stream_info, sizeof(AVISTREAMINFO))))
			break;

		t_tmp.f_frames_per_second = t_avi_stream_info.dwRate / (float)t_avi_stream_info.dwScale;
		t_tmp.f_length = t_avi_stream_info.dwLength / (float)t_avi_stream_info.dwScale;
		t_tmp.n_width = t_avi_stream_info.rcFrame.right - t_avi_stream_info.rcFrame.left;
		t_tmp.n_height = t_avi_stream_info.rcFrame.bottom - t_avi_stream_info.rcFrame.top;
		t_tmp.n_color_planes = -1; // not initialized yet
		t_tmp.n_length = AVIStreamLength(t_tmp.p_avi_stream);
		// calc some info

		long n_size;
		if(AVIStreamReadFormat(t_tmp.p_avi_stream, t_avi_stream_info.dwStart, NULL, &n_size) == 0) {
			BITMAPINFOHEADER *p_format_buffer = (BITMAPINFOHEADER*)malloc(n_size);
			if(!p_format_buffer)
				break;
			if(AVIStreamReadFormat(t_tmp.p_avi_stream, t_avi_stream_info.dwStart, p_format_buffer, &n_size) == 0) {
				BITMAPINFOHEADER t_format;

				memset(&t_format, 0, sizeof(BITMAPINFOHEADER));
				t_format.biSize = sizeof(BITMAPINFOHEADER);
				t_format.biWidth = p_format_buffer->biWidth;
				t_format.biHeight = p_format_buffer->biHeight;
				t_format.biPlanes = 1;
				t_format.biBitCount = 24 + (b_rgba != false) * 8;
				// ja pocitam se 24 (pri kresleni mam GL_BGR), tak tu mam 24, ale ty asi
				// budes chtit 32, aby se to dalo nastrkat do imagestruct. pokud to jen
				// budes rozdelovat na rgb roviny a posilat do ripacu, tak je to jedno
				t_format.biCompression = BI_RGB; // no compression
				// this is what we want

				t_tmp.p_avi_frame = AVIStreamGetFrameOpen(t_tmp.p_avi_stream, &t_format);
				// try to force windows to decompress video for us

				if(t_tmp.p_avi_frame == NULL &&
				   (t_tmp.p_avi_frame = AVIStreamGetFrameOpen(t_tmp.p_avi_stream, NULL)) == NULL)
					break;
				// try at least to show us data
			}
			free(p_format_buffer);
		}

		BITMAPINFO *p_bitmap;
		if(FAILED(p_bitmap = (BITMAPINFO*)AVIStreamGetFrame(t_tmp.p_avi_frame, t_avi_stream_info.dwStart)) || !p_bitmap)
			break;

		assert(t_tmp.n_width == p_bitmap->bmiHeader.biWidth);
		assert(t_tmp.n_height == p_bitmap->bmiHeader.biHeight);
		t_tmp.n_width = p_bitmap->bmiHeader.biWidth;
		t_tmp.n_height = p_bitmap->bmiHeader.biHeight;
		t_tmp.n_color_planes = p_bitmap->bmiHeader.biBitCount;

		TAVI_Stream *p_stream;

		if((p_stream = (TAVI_Stream*)malloc(sizeof(TAVI_Stream))) == NULL)
			break;

		*p_stream = t_tmp;

		return p_stream;
	}

	if(t_tmp.p_avi_frame != NULL)
		AVIStreamGetFrameClose(t_tmp.p_avi_frame);
	if(t_tmp.p_avi_stream != NULL)
		AVIStreamRelease(t_tmp.p_avi_stream);
    AVIFileExit();

	return NULL;
}

void AVI_Free(TAVI_Stream *p_avi)
{
    AVIStreamGetFrameClose(p_avi->p_avi_frame);
    AVIStreamRelease(p_avi->p_avi_stream);
    AVIFileExit();
	free(p_avi);
}

int AVI_GetFrame_t(TAVI_Stream *p_avi, float f_time, unsigned char *p_image, int n_channels, int n_bits_channel)
{
	return AVI_GetFrame_f(p_avi, (int)(f_time * p_avi->f_frames_per_second + .5f), p_image, n_channels, n_bits_channel);
}

int AVI_GetFrame_f(TAVI_Stream *p_avi, long n_frame, unsigned char *p_image, int n_channels, int n_bits_channel)
{
	assert(p_image);
	assert(p_avi);

	n_frame %= p_avi->n_length;
	// budeme opakovat porad dokola

    BITMAPINFO *p_bitmap;
    if(FAILED(p_bitmap = (BITMAPINFO*)AVIStreamGetFrame(p_avi->p_avi_frame, n_frame)) || !p_bitmap)
        return false;

	unsigned char *p_buffer = (unsigned char*)p_bitmap->bmiColors;

	if(p_bitmap->bmiHeader.biPlanes != 1)
		return false;

	int n_bmp_channels = p_bitmap->bmiHeader.biBitCount / 8;

	if(n_bits_channel == 8) {
		switch(n_bmp_channels) {
		case 3:
			if(n_channels == 3) {
				memcpy(p_image, p_buffer, p_avi->n_width * p_avi->n_height * 3 * sizeof(char));
			} else if(n_channels == 4) {
				for(int i = 0, t = 0, q = 0; i < p_avi->n_width * p_avi->n_height; i ++, t += 3, q += 4) {
					p_image[q + 0] = p_buffer[t + 0];
					p_image[q + 1] = p_buffer[t + 1];
					p_image[q + 2] = p_buffer[t + 2];
					p_image[q + 3] = 0xff;
				}
			} else
				return false;
			break;
		case 4:
			if(n_channels == 3) {
				for(int i = 0, t = 0, q = 0; i < p_avi->n_width * p_avi->n_height; i ++, t += 3, q += 4) {
					p_image[t + 0] = p_buffer[q + 0];
					p_image[t + 1] = p_buffer[q + 1];
					p_image[t + 2] = p_buffer[q + 2];
				}
			} else if(n_channels == 4) {
				memcpy(p_image, p_buffer, p_avi->n_width * p_avi->n_height * 4 * sizeof(char));
			} else
				return false;
			break;
		default:
			return false;
		}
	} else
		return false;

	return true;
}
