/*
								+----------------------------------+
								|                                  |
								| *** DigiLib OpenGL extension *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|         DigiLib_Ext.cpp          |
								|                                  |
								+----------------------------------+
*/

#if defined(_MSC_VER) && !defined(__MWERKS__)
#pragma warning(disable:4786)
#endif
// disable warning C4786 "identifier was truncated to '255' characters in the debug information"

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <gl/gl.h>
#include <gl/glut.h>
#include <gl/glext.h>
#include <algorithm>
#include <stdlib.h>
#include <string.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLDrv.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/VertexBufferObject.h"
#include "DigiLib_Ext.h"
#include "RenderBuffer_Manager.h"
#include "Shader_Manager.h"
#include "OpenGL20Logo.h"
#include "ShadersSrc.h"

int CGLNamedRenderBuffer_FBO::m_n_id_space = 0; // id counter for render-buffers
// RenderBuffer_Manager

int CGLNamedShaderObject::m_n_id_space = 0; // id counter for shaders
std::vector<CShaderFactory::TShaderSource> CShaderFactory::m_shader_source_list; // list of shader sources
// Shader_Manager

/*
 *	struct TImageStructInfo
 *		- structure, containing associated data for ImageStruct
 */
struct TImageStructInfo {
	ImageStruct *p_image_struct;
	bool b_current_on_server;
	CGLTexture_2D *p_texture;

	bool b_pixel_buffer_object;
	CGLPixelBufferObject *p_pbo; // contains image, mapped to system memory

	void Delete()
	{
		if(p_texture) {
			delete p_texture;
			p_texture = 0;
		}
		if(p_pbo) {
			if(p_pbo->b_Mapped())
				p_pbo->UnMap();
			delete p_pbo;
			p_pbo = 0;
		}
		if(p_image_struct && b_pixel_buffer_object) {
			free(p_image_struct); // allocated using malloc() in p_Create_GL_Image()
			p_image_struct = 0;
		}
	}
};

static bool b_pixel_buffer_supported = false;
static bool b_fence_supported = false;
static bool b_fbo_supported = false;
static bool b_imaging_supported = false;
static bool b_gl_images_supported = false;
// opengl capabilities

static std::vector<TImageStructInfo> imagestruct_list;
// list of textures, bound to imagestructs

static CGLState *p_state = 0;
// opengl state cache

static bool b_auto_uploads = true;
static bool b_auto_downloads = true;
// automatic uploads / downloads

CGLRenderBuffer_Manager render_buffer_manager;
// render buffer manager object

CGLShader_Manager shader_manager;
// shader manager object

/*
 *								--- function objects for imagestruct_list ---
 */

class CFindImageStructInfo {
protected:
	const ImageStruct *m_p_image_struct;

public:
	__inline CFindImageStructInfo(const ImageStruct *p_image_struct)
		:m_p_image_struct(p_image_struct)
	{}

	__inline bool operator ()(const TImageStructInfo &r_info)
	{
		return r_info.p_image_struct == m_p_image_struct;
	}
};

class CDeleteImageStructInfo {
public:
	__inline void operator ()(TImageStructInfo &r_info)
	{
		r_info.Delete();
	}
};

/*
 *								--- ~function objects for imagestruct_list ---
 */

/*
 *								--- win32 OpenGL ---
 */

#ifdef WIN32
static HWND h_wnd;
static CGLDriver *p_gl_driver = 0;

bool Create_Window(const char *p_s_name, int n_width, int n_height)
{
	WNDCLASSEX t_wnd_class;
	t_wnd_class.cbSize = sizeof(WNDCLASSEX);
	t_wnd_class.style = CS_HREDRAW | CS_VREDRAW;
	t_wnd_class.lpfnWndProc = DefWindowProc;
	t_wnd_class.cbClsExtra = 0;
	t_wnd_class.cbWndExtra = 0;
	t_wnd_class.hInstance = GetModuleHandle(NULL);
	t_wnd_class.hIcon = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	t_wnd_class.hCursor = LoadCursor(GetModuleHandle(NULL), IDC_ARROW);
	t_wnd_class.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	t_wnd_class.lpszMenuName = 0;
	t_wnd_class.lpszClassName = p_s_name;
	t_wnd_class.hIconSm = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	if(!RegisterClassEx(&t_wnd_class))
		return false;
	// create window class

	RECT t_rect;
	t_rect.left = t_rect.top = 0;
	t_rect.right = n_width;
	t_rect.bottom = n_height;
	if(!AdjustWindowRect(&t_rect, WS_BORDER | WS_CAPTION, false))
		return false;
	// calculate window size to contain client area with desired resolution

	if(!(h_wnd = CreateWindow(p_s_name, p_s_name,
	   WS_BORDER | WS_CAPTION, CW_USEDEFAULT, CW_USEDEFAULT,
	   t_rect.right - t_rect.left, t_rect.bottom - t_rect.top,
	   NULL, NULL, GetModuleHandle(NULL), NULL)))
		return false;
	// create window

	ShowWindow(h_wnd, SW_SHOW);
	UpdateWindow(h_wnd);
	// show & update window

	return true;
}

#endif //WIN32

/*
 *								--- ~win32 OpenGL ---
 */

bool Init_OpenGL() // vraci true, pokud se podarilo. pod linuxem pouziva glut, pod windows pouzije COpenGLDriver z berLame
{
	if(p_state)
		return true;

	if(!Register_Shaders())
		return false;
	// register shaders to CShaderFactory

#ifdef WIN32
	if(!Create_Window("DigiLib_Ext", 150, 57))
		return false;
	// create window

	if(!(p_gl_driver = new CGLDriver()))
		return false;
	if(!p_gl_driver->Init(h_wnd, 150, 57, 24, 0)) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// init open-gl

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext()))) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// create state guard
#else //WIN32
	// todo -  linux opengl (init glut)

	int n_arg_num = 1;
	char *p_s_arg = "no-arg";
	glutInit(&n_arg_num, &p_s_arg);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
    glutInitWindowSize(150, 57);
    glutCreateWindow("DigiLib_Ext");

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext()))) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// create state guard
#endif //WIN32

	b_fbo_supported = CGLRenderBuffer_FBO::b_Supported(); // gets function pointers as well
	CGLExtensionHandler::n_GetGL12FuncPointers();
	CGLExtensionHandler::n_GetGL13FuncPointers();
	CGLExtensionHandler::n_GetMultitextureFuncPointers(); // opengl1.3 as well
	CGLExtensionHandler::n_GetGL14FuncPointers();
	CGLExtensionHandler::n_GetGL15FuncPointers();
	CGLExtensionHandler::n_GetGL20FuncPointers();
	b_pixel_buffer_supported = CGLPixelBufferObject::b_Supported(); // gets function pointers as well
	b_gl_images_supported = CGLPixelBufferObject::b_Supported();
	b_fence_supported = CGLExtensionHandler::b_SupportedExtension("GL_NV_fence") &&
		!CGLExtensionHandler::n_GetFenceNVFuncPointers();
	b_imaging_supported = CGLExtensionHandler::b_SupportedExtension("GL_ARB_imaging");
	// get function pointers

	unsigned __int32 *p_gl_logo;
	if((p_gl_logo = COpenGL20Logo::p_DecompressImage())) {
		/*glDrawPixels(n_opengl_logo_2_0_sm_jpg_width, n_opengl_logo_2_0_sm_jpg_height,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);*/ // upside-down
		CGLTexture_2D opengl_tex(p_state, COpenGL20Logo::n_Width(),
			COpenGL20Logo::n_Height(), GL_RGBA8, false, 0, GL_TEXTURE_2D,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);
		delete[] p_gl_logo;

		p_state->DisableDepthTest();
		p_state->DisableCullFace();
		p_state->Color3f(1, 1, 1);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		opengl_tex.Bind_Enable(p_state);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 1);
		glVertex2f(-1, -1);
		glTexCoord2f(0, 0);
		glVertex2f(-1,  1);
		glTexCoord2f(1, 0);
		glVertex2f( 1,  1);
		glTexCoord2f(1, 1);
		glVertex2f( 1, -1);
		glEnd();

#ifdef WIN32
		p_gl_driver->Blit();
#else //WIN32
		glutSwapBuffers();
#endif //WIN32
	}
	// load OpenGL logo

	return true;
}

bool Free_OpenGL_Objects() // odalokuje vsechny shadery, textury, buffery
{
	std::for_each(imagestruct_list.begin(), imagestruct_list.end(), CDeleteImageStructInfo());
	imagestruct_list.clear();
	// delete associated images

	render_buffer_manager.Clear();
	// delete all render-buffers

	shader_manager.Clear();
	// delete all shaders

	return true;
}

bool Shutdown_OpenGL() // ukonci OpenGL
{
	if(!p_state)
		return true;

	bool b_result = Free_OpenGL_Objects();

	delete p_state;
	p_state = 0;

#ifdef WIN32
	b_result &= p_gl_driver->Shutdown();
	delete p_gl_driver;
	// shutdown open-gl

	DestroyWindow(h_wnd);
	// destroy window
#else
	// t_odo - linux opengl (shutdown glut)
	// no function to shutdown glut :(
#endif

	return b_result;
}

bool b_NPOTTexturesSupported() // vraci true, pokud OpenGL podporuje textury, jejichz rozmer neni mocnina dvou
{
	return CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_non_power_of_two");
}

bool b_FloatTexturesSupported() // vraci true, pokud OpenGL podporuje textury s float (IEEE754 32-bit) pixely; je k dispozici i rychlejsi 16 bit format, jez ma ovsem mensi rozsah (+-65030) a pri presunech z a do pameti se musi pouzivat konverze, takze se mu vyhybame
{
	return CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_float");
}

bool b_FramebufferObjectSupported() // vraci true, pokud OpenGL podporuje framebuffer objekty. bez nich nejde renderovat primo do textury, coz prakticky znemoznuje jakekoliv univerzalnejsi zpracovani obrazu
{
	return b_fbo_supported/*CGLRenderBuffer_FBO::b_Supported()*/;
}

int n_MaxTextureSize()
{
	GLint n_size;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &n_size);
	return n_size;
}

int n_PixelSize(int n_pixel_type)
{
	switch(n_pixel_type) {
	case Image8:
	case Image8Linear:
		return sizeof(unsigned char);
	case Image16:
	case Image16Linear:
		return sizeof(short);
	case Image32:
	case Image32Linear:
		return sizeof(long);
	case ImageFloat:
	case ImageFloatLinear:
		return sizeof(float);
	case ImageRGB:
	case ImageRGBLinear:
		return sizeof(unsigned long);
	case ImageComplex:
	case ImageComplexLinear:
		return 2 * sizeof(float);
	default:
		return 0;
	}
}

bool Get_GL_PixelType(int n_pixel_type, int &r_n_fmt, int &r_n_data_type, int &r_n_internal_fmt)
{
	r_n_fmt = GL_LUMINANCE; // 1 channel only, converts to red, 0, 0 when displayed
	r_n_data_type = GL_UNSIGNED_BYTE;

	switch(n_pixel_type) {
	case ImageRGB:
	case ImageRGBLinear:
		r_n_fmt = GL_BGRA;
		r_n_internal_fmt = GL_RGBA;
		break;
	case Image8:
	case Image8Linear:
		//r_n_data_type = GL_UNSIGNED_BYTE;
		r_n_internal_fmt = GL_LUMINANCE8;
		break;
	case Image16:
	case Image16Linear:
		r_n_data_type = GL_SHORT;
		r_n_internal_fmt = GL_LUMINANCE16;
		break;
	case Image32:
	case Image32Linear:
		r_n_data_type = GL_INT;
		r_n_internal_fmt = GL_LUMINANCE16; // beware rounding !!!
		break;
	case ImageFloat:
	case ImageFloatLinear:
		r_n_data_type = GL_FLOAT;
		r_n_internal_fmt = GL_LUMINANCE32F_ARB;
		break;
	case ImageComplex:
	case ImageComplexLinear:
		r_n_fmt = GL_LUMINANCE_ALPHA; // two - channel luminance-alpha image
		r_n_data_type = GL_FLOAT;
		r_n_internal_fmt = GL_LUMINANCE_ALPHA32F_ARB;
		break;
	default:
		return false;
	}

	return true;
}

bool b_FastTransfer(const ImageStruct *p_image) // vraci true, pokud se obrazek bude z/do OpenGL kopirovat rychle (tzn. jednim blokovym presunem)
{
	int n_pixel_size = n_PixelSize(p_image->PixelType);

	if(abs(p_image->XOffset) > 4 || abs(p_image->XOffset) < 1)
		return false;
	// pixels are interleaved

	if(p_image->YOffset != n_pixel_size * p_image->XSize &&
	   p_image->YOffset != -n_pixel_size * p_image->XSize)
	    return false;
	// scanlines are interleaved

	return true;
}

CGLTexture_2D *p_Create_ImageStruct_Texture(const ImageStruct *p_image)
{
	int n_format;
	int n_data_type;
	int n_internal_format;

	if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
		return 0;

	void *p_raster = p_image->Raster;
	int n_y_offset;
	if(p_image->YOffset >= 0)
		n_y_offset = p_image->YOffset;
	else {
		n_y_offset = -p_image->YOffset;
		p_raster = (void*)((int)p_raster - n_y_offset * p_image->YSize);
	}
	int n_x_offset;
	if(p_image->XOffset >= 0)
		n_x_offset = p_image->XOffset;
	else {
		n_x_offset = -p_image->XOffset;
		p_raster = (void*)((int)p_raster - n_x_offset * p_image->XSize);
	}
	// find upper-left corner of image

	if(n_y_offset != n_x_offset * p_image->XSize) {
		_ASSERTE(n_y_offset > n_x_offset * p_image->XSize); // y-offset must be grater than 
		glPixelStorei(GL_UNPACK_ROW_LENGTH, n_y_offset);
	} else
		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	// set proper scanline interleaving

	if(n_x_offset >= 1 && n_x_offset <= 4)
		glPixelStorei(GL_UNPACK_ALIGNMENT, n_x_offset);
	else {
		CGLTexture_2D *p_texture;
		if(!(p_texture = new CGLTexture_2D(p_state, p_image->XSize, p_image->YSize,
		   n_internal_format, false, 0, GL_TEXTURE_2D)))
			return 0;
		if(!p_texture->b_Status()) {
			delete p_texture;
			return 0;
		}
		// create empty texture

		p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		// clamp to edge

		int n_x = 0;
		for(void *p_end = (void*)((int)p_raster + p_image->XSize * n_x_offset);
		   p_raster != p_end; p_raster = (void*)((int)p_raster + n_x_offset), ++ n_x) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, n_x, 0, 1, p_image->YSize,
				n_format, n_data_type, p_raster);
		}
		// copy image as vertical scanlines (slow!)

		return p_texture;
	}
	// set proper pixel alignment

	CGLTexture_2D *p_texture;
	if(!(p_texture = new CGLTexture_2D(p_state, p_image->XSize, p_image->YSize,
	   n_internal_format, false, 0, GL_TEXTURE_2D, n_format, n_data_type, p_raster)))
		return 0;
	if(!p_texture->b_Status()) {
		delete p_texture;
		return 0;
	}
	// create texture

	p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	// clamp to edge

	return p_texture;
}

bool b_HasTexture(const ImageStruct *p_image) // vraci true, pokud obrazek ma asociovanou nejakou texturu
{
	return std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
		CFindImageStructInfo(p_image)) != imagestruct_list.end();
}

bool Upload_ImageStruct(ImageStruct *p_image) // nahraje obrazek do graficke karty
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) == imagestruct_list.end()) {
	    imagestruct_list.reserve(imagestruct_list.size() + 1);
		if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
			return false;
		// no more place in the list

		TImageStructInfo t_temp;
		t_temp.b_current_on_server = true;
		t_temp.b_pixel_buffer_object = false;
		t_temp.p_image_struct = p_image;
		t_temp.p_pbo = 0;
		if(!(t_temp.p_texture = p_Create_ImageStruct_Texture(p_image)))
			return false;
		// create texture

		imagestruct_list.push_back(t_temp);
		p_image_info = imagestruct_list.end() - 1;
		// put it into list, update iterator
	}
	// make sure there is associated texture

	// todo - upload

	return true;
}

bool Download_ImageStruct(ImageStruct *p_image) // stahne obrazek z graficke karty
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) != imagestruct_list.end()) {
		// todo - download

	    return true;
	}

	return false;
	// has no associated texture
}

bool b_AsyncSupported() // vrati true, pokud jsou podporovany asynchronni prenosy
{
	return b_pixel_buffer_supported;//CVertexBufferObject::b_Supported() && CGLPixelBufferObject::b_Supported();
}

bool b_FenceSupported() // vrati true, pokud je podporovan NV_FENCE (zjistovani dokonceni operaci, ati stale nema ekvivalent)
{
	return b_fence_supported;//CGLExtensionHandler::b_SupportedExtension("GL_NV_fence");
}

bool Upload_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id); // asynchronne nahraje obrazek do graficke karty; do p_fence_id nahraje id plotu
bool Download_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id); // asynchronne stahne obrazek do graficke karty; do p_fence_id nahraje id plotu

bool b_FenceFinished(unsigned int n_fence_id) // vrati true, pokud je operace za niz nasledoval plot n_fence_id dokoncena; jinak false
{
	if(!b_fence_supported) {
		glFlush();
		glFinish();
		return true;
	}
	return glTestFenceNV(n_fence_id) != 0;
}

bool WaitFence(unsigned int n_fence_id) // ceka dokud plot n_fence_id neskonci
{
	if(!b_fence_supported) {
		glFlush();
		glFinish();
		return true;
	}
	if(!glIsFenceNV(n_fence_id)) {
		glFlush();
		glFinish();
		return false;
	}
	glFinishFenceNV(n_fence_id);
	return true;
}

void Enable_AutoUploads() // povoli automaticke uploady (default; obrazek se zkopiruje ze sve pameti pokazde kdyz se s nim provadi nejaka operace)
{
	b_auto_uploads = true;
}

void Enable_AutoDownloads() // povoli automaticke downloady (default; obrazek se zkopiruje z karty pokazde kdyz se do nej provadi nejaka operace)
{
	b_auto_downloads = true;
}

void Disable_AutoUploads()
{
	b_auto_uploads = false;
}

void Disable_AutoDownloads() // zakaze automaticke uploady / downloady
{
	b_auto_downloads = false;
}

bool Set_DirtyFlag(const ImageStruct *p_image, bool b_current_on_server) // nastavi priznak modifikace obrazu. b_current_on_server rika zda je obrazek aktualni na graficke karte (true) nebo v systemove pameti (false); prvni funkce jez s obrazkem bude operovat potom muze provest potrebnou kopii a vynulovat priznak (tzn. bud nejaka operace nad obrazkem nebo [Up/Down]load[_Async]_ImageStruct())
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) != imagestruct_list.end()) {
		(*p_image_info).b_current_on_server = b_current_on_server; // set attrib
	    return true;
	}

	return b_current_on_server == false;
	// in case we wanted to invalidate server-side copy, we haven't failed
	// (has no associated texture)
}

bool b_GL_ImagesSupported() // vrati true pokud jsou podporovany obrazky, lezici v graficke pameti mapovane do systemove ram
{
	return b_gl_images_supported;
}

bool b_IsGL_Image(const ImageStruct *p_image) // vrati true, pokud je dany imagestruct takovym obrazkem
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) != imagestruct_list.end())
	    return (*p_image_info).b_pixel_buffer_object;
	return false;
}

bool b_GL_Image_RelocTest() // vrati true, pokud se adresy dat obrazku mezi jednotlivymi operacemi meni (pokusi se to vyzkouset, negativni vysledek nemusi odpovidat pravde)
{
	// todo
	return true;
}

ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format) // vytvori mapovany obrazek o danych rozmerech a formatu (vzdycky je to linearni blok pixelu; nepodporuje tedy formaty jez nekonci _linear)
{
	int n_internal_format, n_fmt, n_data;
	if(!Get_GL_PixelType(n_format, n_fmt, n_data, n_internal_format))
		return 0;
	// get internal format

	imagestruct_list.reserve(imagestruct_list.size() + 1);
	if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
		return 0;
	// no more place in the list

	TImageStructInfo t_temp;
	t_temp.b_current_on_server = false;
	t_temp.b_pixel_buffer_object = true;
	if(!(t_temp.p_texture = new CGLTexture_2D(p_state, n_width, n_height, n_internal_format, false)))
		return 0;
	if(!t_temp.p_texture->b_Status()) {
		delete t_temp.p_texture;
		return 0;
	}
	t_temp.p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	t_temp.p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	// create texture, set clamp to edge

	if(!(t_temp.p_pbo = new CGLPixelBufferObject)) {
		delete t_temp.p_texture;
		return 0;
	}
	if(!t_temp.p_pbo->Alloc(n_PixelSize(n_format) * n_width * n_height)) {
		delete t_temp.p_texture;
		delete t_temp.p_pbo;
		return 0;
	}
	// create pixel-buffer object to be associated with texture // todo - don't remember how to bind

	ImageStruct *p_image_struct;
	if((p_image_struct = (ImageStruct*)malloc(sizeof(ImageStruct))) == NULL) {
		delete t_temp.p_texture;
		delete t_temp.p_pbo;
		return 0;
	}
	p_image_struct->PixelType = n_format;
	p_image_struct->XOffset = n_PixelSize(n_format);
	p_image_struct->XSize = n_width;
	p_image_struct->YOffset = n_width * n_PixelSize(n_format);
	p_image_struct->YSize = n_height;
	p_image_struct->Raster = (unsigned char*)t_temp.p_pbo->p_Map(GL_READ_WRITE_ARB);
	p_image_struct->Optional.Regular.Free = Destroy_GL_Image;
	// fill imagestruct

	t_temp.p_image_struct = p_image_struct;

	imagestruct_list.push_back(t_temp);
	// put it into list

	return p_image_struct;
}

void Destroy_GL_Image(void *p_gl_image) // zrusi mapovany obrazek (sem ukazuje pointer Free v ImageStruct)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo((ImageStruct*)p_gl_image))) != imagestruct_list.end()) {
	    (*p_image_info).Delete();
		imagestruct_list.erase(p_image_info, p_image_info + 1);
	}
	// find and destroy TImageStructInfo and OpenGL objects

	//free(p_gl_image); // (*p_image_info).Delete(); line handles it
	// free imagestruct
}

bool Bind_SrcImage(ImageStruct *p_image)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) == imagestruct_list.end()) {
	    imagestruct_list.reserve(imagestruct_list.size() + 1);
		if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
			return false;
		// no more place in the list

		TImageStructInfo t_temp;
		t_temp.b_current_on_server = true;
		t_temp.b_pixel_buffer_object = false;
		t_temp.p_image_struct = p_image;
		t_temp.p_pbo = 0;
		if(!(t_temp.p_texture = p_Create_ImageStruct_Texture(p_image)))
			return false;
		// create texture

		imagestruct_list.push_back(t_temp);
		p_image_info = imagestruct_list.end() - 1;
		// put it into list, update iterator
	}
	// make sure there is associated texture

	_ASSERTE((*p_image_info).p_texture);
	(*p_image_info).p_texture->Bind_Enable(p_state);
	// bind texture

	if((*p_image_info).b_pixel_buffer_object) {
		_ASSERTE((*p_image_info).p_pbo);
		if((*p_image_info).p_pbo->b_Mapped())
			(*p_image_info).p_pbo->UnMap();
	}
	// destroy memory mapping

	if(!b_auto_uploads && (*p_image_info).b_current_on_server)
		return true;
	// in case image is current on server, do nothing

	if((*p_image_info).b_pixel_buffer_object) {
		(*p_image_info).p_pbo->BindAs(GL_PIXEL_UNPACK_BUFFER_ARB);

		int n_format, n_data_type, n_internal_format;
		Get_GL_PixelType((*p_image_info).p_image_struct->PixelType,
			n_format, n_data_type, n_internal_format);

		glTexSubImage2D(GL_TEXTURE_2D,
					0, // LOD pyramid level
					0, // x-offset
					0, // y-offset
					(*p_image_info).p_texture->n_Width(), // width
					(*p_image_info).p_texture->n_Height(), // height
					n_format,
					n_data_type,
					(*p_image_info).p_pbo->p_OffsetPointer(0));
		// tell OpenGL to source texture from PBO

		CGLBufferObjectUnBinder::UnBind(GL_PIXEL_UNPACK_BUFFER_ARB);
	} else {
		// todo - upload traditional imagestruct
	}

	return true;
}

bool Release_SrcImage(ImageStruct *p_image)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image));
	_ASSERTE(p_image_info != imagestruct_list.end()); // there must be associated texture

	if((*p_image_info).b_pixel_buffer_object) {
		_ASSERTE((*p_image_info).p_pbo);
		if(!(p_image->Raster = (unsigned char*)(*p_image_info).p_pbo->p_Map(GL_READ_WRITE_ARB)))
			return false;
	}
	// refresh memory mapping

	return true;
}

static CGLNamedRenderBuffer_FBO *p_cur_render_buffer = 0;

bool Bind_DestImage(ImageStruct *p_image)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	if((p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image))) == imagestruct_list.end()) {
	    imagestruct_list.reserve(imagestruct_list.size() + 1);
		if(imagestruct_list.capacity() < imagestruct_list.size() + 1)
			return false;
		// no more place in the list

		TImageStructInfo t_temp;
		t_temp.b_current_on_server = true;
		t_temp.b_pixel_buffer_object = false;
		t_temp.p_image_struct = p_image;
		t_temp.p_pbo = 0;
		if(!(t_temp.p_texture = p_Create_ImageStruct_Texture(p_image)))
			return false;
		// create texture

		imagestruct_list.push_back(t_temp);
		p_image_info = imagestruct_list.end() - 1;
		// put it into list, update iterator
	}
	// make sure there is associated texture

	if((*p_image_info).b_pixel_buffer_object) {
		_ASSERTE((*p_image_info).p_pbo);
		if((*p_image_info).p_pbo->b_Mapped())
			(*p_image_info).p_pbo->UnMap();
	}
	// destroy memory mapping

	int n_fbo_id = render_buffer_manager.n_UnusedObject_Id((*p_image_info).p_texture, -1);
	if(n_fbo_id == -1) {
		if(!render_buffer_manager.LoadObject((*p_image_info).p_texture, CRenderBufferFactory()))
			return false;
		n_fbo_id = render_buffer_manager.n_UnusedObject_Id((*p_image_info).p_texture, -1);
	}
	_ASSERTE(n_fbo_id != -1);
	// find / create appropriate render-buffer

	p_cur_render_buffer = &render_buffer_manager.r_Object(n_fbo_id);
	render_buffer_manager.ReferenceObject(n_fbo_id, false);
	// get our dreamy render buffer

	glViewport(0, 0, p_cur_render_buffer->n_Width(), p_cur_render_buffer->n_Height());
	// set the right viewport

	_ASSERTE((*p_image_info).p_texture);
	(*p_image_info).p_texture->Disable(p_state);
	// disable texture

	p_cur_render_buffer->Bind();
	p_cur_render_buffer->BindColorTexture_2D(*(*p_image_info).p_texture, GL_TEXTURE_2D, 0);
	// bind render-buffer and bind texture for rendering

	return true;
}

bool Release_DestImage(ImageStruct *p_image)
{
	std::vector<TImageStructInfo>::iterator p_image_info;
	p_image_info = std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
	   CFindImageStructInfo(p_image));
	_ASSERTE(p_image_info != imagestruct_list.end()); // there must be associated texture

	_ASSERTE(p_cur_render_buffer);
	p_cur_render_buffer->CopyTextureData(true);
	p_cur_render_buffer->Release();
	// release textures and render buffer (note no copying is actually made,
	// all rendering is made directly to the textures, calling it to release textures only)

	render_buffer_manager.ReturnObject(p_cur_render_buffer->n_Id(), false);
	p_cur_render_buffer = 0;
	// return render-buffer-object back to circulation

	if(b_auto_downloads) {
		if((*p_image_info).b_pixel_buffer_object) {
			(*p_image_info).p_pbo->Bind();

			int n_format, n_data_type, n_internal_format;
			Get_GL_PixelType((*p_image_info).p_image_struct->PixelType,
				n_format, n_data_type, n_internal_format);

			(*p_image_info).p_texture->Bind_Enable(p_state);
			glGetTexImage(GL_TEXTURE_2D,
						0, // LOD pyramid level
						n_format,
						n_data_type,
						(*p_image_info).p_pbo->p_OffsetPointer(0));
			// fast block copy in graphics card (texture to buffer)
		} else {
			// todo - download traditional imagestruct
		}
	}

	if((*p_image_info).b_pixel_buffer_object) {
		_ASSERTE((*p_image_info).p_pbo);
		(*p_image_info).p_pbo->Bind();
		if(!(p_image->Raster = (unsigned char*)(*p_image_info).p_pbo->p_Map(GL_READ_WRITE_ARB)))
			return false;
	}
	// refresh memory mapping

	return true;
}

static bool b_error_level = false;

bool b_ErrorLevel()
{
	if(b_error_level) {
		b_error_level = false;
		return true;
	}
	return false;
}

void ImageImageAdd(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	p_state->ActiveTextureUnit(1);
	if(!Bind_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// bind destination image as render-buffer

	if(!Bind_SrcImage(p_src_image1)) {
		b_error_level = true;
		return;
	}
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);
	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_src_image0)) {
		b_error_level = true;
		return;
	}
	// bind source images as textures

	p_state->DisableDepthTest();
	p_state->DisableFog();
	p_state->DisableLighting();
	p_state->DisableAllClipPlanes();
	p_state->DisableAlphaTest();
	p_state->DisableFragmentProgram();
	p_state->DisableStencilTest();
	p_state->DisableVertexProgram();
	p_state->Disable_STRQ_Texgen();
	p_state->DisableBlend();
	p_state->Color3f(1, 1, 1);
	// set required GL state

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	glBegin(GL_QUADS);
	glVertex2f(-1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 0);
	glVertex2f(-1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 0);
	glVertex2f( 1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 1);
	glVertex2f( 1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 1);
	glEnd();
	// render fullscreen quad

	if(!Release_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// unbind render-buffer

	p_state->ActiveTextureUnit(1);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); // return default
	if(!Release_SrcImage(p_src_image1)) {
		b_error_level = true;
		return;
	}
	p_state->ActiveTextureUnit(0);
	if(!Release_SrcImage(p_src_image0)) {
		b_error_level = true;
		return;
	}
	// refresh memory mapping for source images if necessary
}

void ImageImageMul(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	p_state->ActiveTextureUnit(1);
	if(!Bind_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// bind destination image as render-buffer

	if(!Bind_SrcImage(p_src_image1)) {
		b_error_level = true;
		return;
	}
	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_src_image0)) {
		b_error_level = true;
		return;
	}
	// bind source images as textures

	p_state->DisableDepthTest();
	p_state->DisableFog();
	p_state->DisableLighting();
	p_state->DisableAllClipPlanes();
	p_state->DisableAlphaTest();
	p_state->DisableFragmentProgram();
	p_state->DisableStencilTest();
	p_state->DisableVertexProgram();
	p_state->Disable_STRQ_Texgen();
	p_state->DisableBlend();
	p_state->Color3f(1, 1, 1);
	// set required GL state

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	glBegin(GL_QUADS);
	glVertex2f(-1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 0);
	glVertex2f(-1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 0);
	glVertex2f( 1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 1);
	glVertex2f( 1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 1);
	glEnd();
	// render fullscreen quad

	if(!Release_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// unbind render-buffer

	p_state->ActiveTextureUnit(1);
	if(!Release_SrcImage(p_src_image1)) {
		b_error_level = true;
		return;
	}
	p_state->ActiveTextureUnit(0);
	if(!Release_SrcImage(p_src_image0)) {
		b_error_level = true;
		return;
	}
	// refresh memory mapping for source images if necessary
}

/*void ImageImageTestShader(ImageStruct *p_dest_image,
	ImageStruct *p_src_image0, ImageStruct *p_src_image1)
{
	p_state->ActiveTextureUnit(1);
	if(!Bind_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// bind destination image as render-buffer

	if(!Bind_SrcImage(p_src_image1)) {
		b_error_level = true;
		return;
	}
	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_src_image0)) {
		b_error_level = true;
		return;
	}
	// bind source images as textures

	p_state->DisableDepthTest();
	p_state->DisableFog();
	p_state->DisableLighting();
	p_state->DisableAllClipPlanes();
	p_state->DisableAlphaTest();
	p_state->DisableFragmentProgram();
	p_state->DisableStencilTest();
	p_state->DisableVertexProgram();
	p_state->Disable_STRQ_Texgen();
	p_state->DisableBlend();
	p_state->Color3f(1, 1, 1);
	// set required GL state

	const char *p_s_shader_name = "|my_test_shader2";
	// shader name required multiple times below. this is here to avoid typos only

	int n_shader_id;
	if((n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1 &&
	   (!shader_manager.LoadObject(p_s_shader_name, CShaderFactory(p_state)) ||
	   (n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1)) {
		b_error_level = true;
		return;
	}
	CGLNamedShaderObject &r_shader = shader_manager.r_Object(n_shader_id);
	r_shader.Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	glBegin(GL_QUADS);
	glVertex2f(-1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 0);
	glVertex2f(-1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 0);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 0);
	glVertex2f( 1, -1);
	glMultiTexCoord2f(GL_TEXTURE0, 1, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 1, 1);
	glVertex2f( 1,  1);
	glMultiTexCoord2f(GL_TEXTURE0, 0, 1);
	glMultiTexCoord2f(GL_TEXTURE1, 0, 1);
	glEnd();
	// render fullscreen quad

	r_shader.Release(p_state);
	// unbind shader

	if(!Release_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// unbind render-buffer

	p_state->ActiveTextureUnit(1);
	if(!Release_SrcImage(p_src_image1)) {
		b_error_level = true;
		return;
	}
	p_state->ActiveTextureUnit(0);
	if(!Release_SrcImage(p_src_image0)) {
		b_error_level = true;
		return;
	}
	// refresh memory mapping for source images if necessary
}*/

void ImageToGreyscale(ImageStruct *p_dest_image, ImageStruct *p_src_image)
{
	if(!Bind_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// bind destination image as render-buffer

	if(!Bind_SrcImage(p_src_image)) {
		b_error_level = true;
		return;
	}
	// bind source image as textures

	p_state->DisableDepthTest();
	p_state->DisableFog();
	p_state->DisableLighting();
	p_state->DisableAllClipPlanes();
	p_state->DisableAlphaTest();
	p_state->DisableFragmentProgram();
	p_state->DisableStencilTest();
	p_state->DisableVertexProgram();
	p_state->Disable_STRQ_Texgen();
	p_state->DisableBlend();
	p_state->Color3f(1, 1, 1);
	// set required GL state

	const char *p_s_shader_name = "|greyscale_shader";
	// shader name required multiple times below. this is here to avoid typos only

	int n_shader_id;
	if((n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1 &&
	   (!shader_manager.LoadObject(p_s_shader_name, CShaderFactory(p_state)) ||
	   (n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1)) {
		b_error_level = true;
		return;
	}
	CGLNamedShaderObject &r_shader = shader_manager.r_Object(n_shader_id);
	r_shader.Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	glBegin(GL_QUADS);
	glVertex2f(-1,  1);
	glTexCoord2f(0, 0);
	glVertex2f(-1, -1);
	glTexCoord2f(1, 0);
	glVertex2f( 1, -1);
	glTexCoord2f(1, 1);
	glVertex2f( 1,  1);
	glTexCoord2f(0, 1);
	glEnd();
	// render fullscreen quad

	r_shader.Release(p_state);
	// unbind shader

	if(!Release_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// unbind render-buffer

	if(!Release_SrcImage(p_src_image)) {
		b_error_level = true;
		return;
	}
	// refresh memory mapping for source image if necessary
}

void ImageColorMatrix(ImageStruct *p_dest_image,
	ImageStruct *p_src_image, const float *p_mat4x4, bool b_transpose)
{
	if(!Bind_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// bind destination image as render-buffer

	if(!Bind_SrcImage(p_src_image)) {
		b_error_level = true;
		return;
	}
	// bind source image as textures

	p_state->DisableDepthTest();
	p_state->DisableFog();
	p_state->DisableLighting();
	p_state->DisableAllClipPlanes();
	p_state->DisableAlphaTest();
	p_state->DisableFragmentProgram();
	p_state->DisableStencilTest();
	p_state->DisableVertexProgram();
	p_state->Disable_STRQ_Texgen();
	p_state->DisableBlend();
	p_state->Color3f(1, 1, 1);
	// set required GL state

	const char *p_s_shader_name = "|color-matrix_shader";
	// shader name required multiple times below. this is here to avoid typos only

	int n_shader_id;
	if((n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1 &&
	   (!shader_manager.LoadObject(p_s_shader_name, CShaderFactory(p_state)) ||
	   (n_shader_id = shader_manager.n_Object_Id(p_s_shader_name, -1)) == -1)) {
		b_error_level = true;
		return;
	}
	CGLNamedShaderObject &r_shader = shader_manager.r_Object(n_shader_id);
	r_shader.p_Find_MatrixParameter("matrix")->SetValue(p_mat4x4, b_transpose); // set matrix value
	r_shader.Bind(p_state);
	// bind shader (lazy compile)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	glBegin(GL_QUADS);
	glVertex2f(-1,  1);
	glTexCoord2f(0, 0);
	glVertex2f(-1, -1);
	glTexCoord2f(1, 0);
	glVertex2f( 1, -1);
	glTexCoord2f(1, 1);
	glVertex2f( 1,  1);
	glTexCoord2f(0, 1);
	glEnd();
	// render fullscreen quad

	r_shader.Release(p_state);
	// unbind shader

	if(!Release_DestImage(p_dest_image)) {
		b_error_level = true;
		return;
	}
	// unbind render-buffer

	if(!Release_SrcImage(p_src_image)) {
		b_error_level = true;
		return;
	}
	// refresh memory mapping for source image if necessary
}

void GetImageHistogram(ImageStruct *p_src_image, int n_histogram_width, unsigned int *p_histogram)
{
	if(!Bind_SrcImage(p_src_image)) {
		b_error_level = true;
		return;
	}
	// bind source image as texture

	GLenum n_pixel_format = (*std::find_if(imagestruct_list.begin(), imagestruct_list.end(),
		CFindImageStructInfo(p_src_image))).p_texture->n_Pixel_Format();
	// find image pixel format

	if(n_pixel_format == GL_BGRA)
		n_pixel_format = GL_RGBA; // BGRA not allowed, no swizzling circuitry here

	p_state->Enable(GL_HISTOGRAM);
	glHistogram(GL_HISTOGRAM, n_histogram_width, n_pixel_format, GL_TRUE);
	//glResetHistogram(GL_HISTOGRAM); // done by glGetHistogram()
	// create and enable histogram

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	glBegin(GL_QUADS);
	glVertex2f(-1,  1);
	glTexCoord2f(0, 0);
	glVertex2f(-1, -1);
	glTexCoord2f(1, 0);
	glVertex2f( 1, -1);
	glTexCoord2f(1, 1);
	glVertex2f( 1,  1);
	glTexCoord2f(0, 1);
	glEnd();
	// render fullscreen quad

	glGetHistogram(GL_HISTOGRAM, GL_TRUE, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, p_histogram);
	// reset histogram as well

	p_state->Disable(GL_HISTOGRAM);
	// disable histogram

	if(!Release_SrcImage(p_src_image)) {
		b_error_level = true;
		return;
	}
}

/*
 *		-end-of-file-
 */
