/*
								+----------------------------------+
								|                                  |
								| *** DigiLib OpenGL extension *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|          DigiLib_Ext.h           |
								|                                  |
								+----------------------------------+
*/

#ifndef __DIGILIB_EXT_INCLUDED
#define __DIGILIB_EXT_INCLUDED

/* -digilib-defs-begin- */

#ifdef INCLUDE_DIGILIB
#include <image.h>
#else //INCLUDE_DIGILIB

typedef unsigned char Image8Value;
typedef short Image16Value;
typedef long Image32Value;
typedef float ImageFloatValue;
typedef unsigned char ImageRGBValue;
typedef unsigned long ImageRGBValueRGB;
typedef struct { float Re; float Im; } ImageComplexValue;

typedef struct ImageStruct {
  unsigned char * Raster;
  char * Info;
  void * BinInfo;
  int BinInfoSize;
  void *ExtraInfo;
  int XSize;
  int YSize;
  int XOffset;
  int YOffset;
  int InfoSize;
  short PixelType;
  short ImageType;
  int References;
  union {
    struct {
      void (* Free)(void *);
    } Regular;
    struct {
      void (* Free)(void *);
      struct ImageStruct * Referenced;
    } Reference;
    struct {
      void (* Free)(void *);
      void (* FreeData)(void *);
      void * Data;
    } External;
  } Optional;
} ImageStruct;

/* Pixel Type Definitions */

#define Image8                 ((short)(0x0010|sizeof(Image8Value)))
#define Image8Linear           ((short)(0x0030|sizeof(Image8Value)))
#define Image16                ((short)(0x0040|sizeof(Image16Value)))
#define Image16Linear          ((short)(0x00C0|sizeof(Image16Value)))
#define Image32                ((short)(0x0100|sizeof(Image32Value)))
#define Image32Linear          ((short)(0x0300|sizeof(Image32Value)))
#define ImageFloat             ((short)(0x0400|sizeof(ImageFloatValue)))
#define ImageFloatLinear       ((short)(0x0C00|sizeof(ImageFloatValue)))
#define ImageRGB               ((short)(0x1000|sizeof(ImageRGBValueRGB)))
#define ImageRGBLinear         ((short)(0x3000|sizeof(ImageRGBValueRGB)))
#define ImageComplex           ((short)(0x4000|sizeof(ImageComplexValue)))
#define ImageComplexLinear     ((short)(0xC000|sizeof(ImageComplexValue)))

#define ImageLinearMask        ((short)(0xAAA0))
#define ImageValueMask         0xF

/* Image Type Definitions */

#define RegularImage           0x01
#define ReferenceImage         0x02
#define ExternalImage          0x04
#define TempImage              0x08

/* Info Definitions */

#define DefaultInfoSize        768

#endif //INCLUDE_DIGILIB

/* -digilibs-defs-end- */

#ifdef __cplusplus
extern "C" {
#endif

bool Init_OpenGL(); // vraci true, pokud se podarilo. pod linuxem pouziva glut, pod windows pouzije COpenGLDriver z berLame
bool Free_OpenGL_Objects(); // odalokuje vsechny shadery, textury, buffery
bool Shutdown_OpenGL(); // ukonci OpenGL

bool b_NPOTTexturesSupported(); // vraci true, pokud OpenGL podporuje textury, jejichz rozmer neni mocnina dvou
bool b_FloatTexturesSupported(); // vraci true, pokud OpenGL podporuje textury s float (IEEE754 32-bit) pixely; je k dispozici i rychlejsi 16 bit format, jez ma ovsem mensi rozsah (+-65030) a pri presunech z a do pameti se musi pouzivat konverze, takze se mu vyhybame
bool b_FramebufferObjectSupported(); // vraci true, pokud OpenGL podporuje framebuffer objekty. bez nich nejde renderovat primo do textury, coz prakticky znemoznuje jakekoliv univerzalnejsi zpracovani obrazu
int n_MaxTextureSize(); // vraci maximalni rozmery (2D) textury

bool b_FastTransfer(const ImageStruct *p_image); // vraci true, pokud se obrazek bude z/do OpenGL kopirovat rychle (tzn. jednim blokovym presunem)
bool b_HasTexture(const ImageStruct *p_image); // vraci true, pokud obrazek ma asociovanou nejakou texturu

bool Upload_ImageStruct(ImageStruct *p_image); // nahraje obrazek do graficke karty
bool Download_ImageStruct(ImageStruct *p_image); // stahne obrazek z graficke karty

bool b_AsyncSupported(); // vrati true, pokud jsou podporovany asynchronni prenosy
bool b_FenceSupported(); // vrati true, pokud je podporovan NV_FENCE (zjistovani dokonceni operaci, ati stale nema ekvivalent)

bool Upload_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id); // asynchronne nahraje obrazek do graficke karty; do p_fence_id nahraje id plotu
bool Download_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id); // asynchronne stahne obrazek do graficke karty; do p_fence_id nahraje id plotu

bool b_FenceFinished(unsigned int n_fence_id); // vrati true, pokud je operace za niz nasledoval plot n_fence_id dokoncena; jinak false
bool WaitFence(unsigned int n_fence_id); // ceka dokud plot n_fence_id neskonci

void Enable_AutoUploads(); // povoli automaticke uploady (default; obrazek se zkopiruje ze sve pameti pokazde kdyz se s nim provadi nejaka operace)
void Enable_AutoDownloads(); // povoli automaticke downloady (default; obrazek se zkopiruje z karty pokazde kdyz se do nej provadi nejaka operace)
void Disable_AutoUploads();
void Disable_AutoDownloads(); // zakaze automaticke uploady / downloady

bool Set_DirtyFlag(const ImageStruct *p_image, bool b_current_on_server); // nastavi priznak modifikace obrazu. b_current_on_server rika zda je obrazek aktualni na graficke karte (true) nebo v systemove pameti (false); prvni funkce jez s obrazkem bude operovat potom muze provest potrebnou kopii a vynulovat priznak (tzn. bud nejaka operace nad obrazkem nebo [Up/Down]load[_Async]_ImageStruct())

bool b_GL_ImagesSupported(); // vrati true pokud jsou podporovany obrazky, lezici v graficke pameti mapovane do systemove ram
bool b_IsGL_Image(const ImageStruct *p_image); // vrati true, pokud je dany imagestruct takovym obrazkem
bool b_GL_Image_RelocTest(); // vrati true, pokud se adresy dat obrazku mezi jednotlivymi operacemi meni (pokusi se to vyzkouset, negativni vysledek nemusi odpovidat pravde)

ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format); // vytvori mapovany obrazek o danych rozmerech a formatu (vzdycky je to linearni blok pixelu; nepodporuje tedy formaty jez nekonci _linear)
void Destroy_GL_Image(void *p_gl_image); // zrusi mapovany obrazek (sem ukazuje pointer Free v ImageStruct)

bool b_ErrorLevel(); // pokud nastala chyba pri provadeni operaci nad obrazem, vraci true a nuluje priznak chyby. jinak vraci false

// mnozina zatim implementovanych funkci

void ImageImageAdd(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);
void ImageImageMul(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

void ImageToGreyscale(ImageStruct *p_dest_image, ImageStruct *p_src_image);
// converts image to greyscale using classical (.299, .587, .144) dot product

void ImageColorMatrix(ImageStruct *p_dest_image,
	ImageStruct *p_src_image, const float *p_mat4x4, bool b_transpose);
// treats each pixel as 4-dimensional vector (RGBA) and multiplies it by a given matrix
// (mat4x4 is color matrix, in case b_transpose is nonzero, it will be transposed before applying)

/*
 *	for example this matrix can be used to convert RGB images to YCbCr colorspace:
 *	(note alpha channel of such an image must contain value 1 (respectively 255) in all pixels)
 *
 *	float p_mat4x4[16] = { // RGB to YCbCr (ITU-R BT.601)
 *		    .299f,     .587f,     .114f,   0, // R' = .299R + .587G + .114B
 *		-.168736f, -.331264f,       .5f, .5f, // G' = .5 - .168736R - .331264G - .5B
 *		      .5f, -.418688f, -.081312f, .5f, // B' = .5 + .5R - .418688G - .081312B
 *		        0,         0,         0,   1  // A' = 1
 *	};
 */

void GetImageHistogram(ImageStruct *p_src_image, int n_histogram_width, unsigned int *p_histogram);
// creates image histogram of p_src_image; n_histogram_width should be positive pover of two,
// p_histogram is array histogram will be copied to (will contain n_histogram_width values)

#ifdef __cplusplus
}
#endif

#endif //__DIGILIB_EXT_INCLUDED
