#include <windows.h>
#include <time.h>
#include <gl/gl.h>
#include <stdio.h>
#include "DigiLib_Ext.h"
#include "AVIStream.h"
#include "../berLame_src/Timer.h"

struct TBmp {
	char n_former_bpp; // former bpp, before conversion to RGBA8
	bool b_grayscale;
	bool b_alpha;
	int n_width, n_height;
	unsigned __int32 *p_buffer;
};

bool Save_RGBLinearImageStruct(const char *p_s_file, const ImageStruct *p_imagestruct)
{
	if(p_imagestruct->PixelType != ImageRGBLinear)
		return false;
	// optimised for ImageRGBLinear images

	FILE *p_fw;
	if(!(p_fw = fopen(p_s_file, "wb")))
		return false;
	// open file

	BITMAPINFOHEADER t_info_header = {0};
	t_info_header.biSize = sizeof(BITMAPINFOHEADER);
	t_info_header.biWidth = p_imagestruct->XSize;
	t_info_header.biHeight = p_imagestruct->YSize;
	t_info_header.biPlanes = 1;
	t_info_header.biBitCount = 24;
	t_info_header.biCompression = BI_RGB;
	t_info_header.biSizeImage = p_imagestruct->XSize * p_imagestruct->YSize * 3;
	t_info_header.biXPelsPerMeter = 96;
	t_info_header.biYPelsPerMeter = 96;

	BITMAPFILEHEADER t_file_header = {0};
	t_file_header.bfType = ('B' | ('M' << 8));
	t_file_header.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	t_file_header.bfSize = t_file_header.bfOffBits + t_info_header.biSizeImage;
	// create header

	if(fwrite(&t_file_header, sizeof(BITMAPFILEHEADER), 1, p_fw) != 1 ||
	   fwrite(&t_info_header, sizeof(BITMAPINFOHEADER), 1, p_fw) != 1) {
		fclose(p_fw);
		return false;
	}
	// write header

	int n_scanline_padd = (3 * p_imagestruct->XSize) & 3;
	// length of scanline padding (padd to 4-byte multiples)

	for(const unsigned char *p_raster = p_imagestruct->Raster, *p_end = p_imagestruct->Raster +
	   (p_imagestruct->XSize * p_imagestruct->YSize * 4); p_raster != p_end;) {
		for(const unsigned char *p_scanline_end = p_raster + p_imagestruct->XSize * 4;
		   p_raster != p_scanline_end; p_raster += 2) {
			if(fwrite(p_raster, sizeof(unsigned char), 1, p_fw) != 1 ||
			   fwrite(++ p_raster, sizeof(unsigned char), 1, p_fw) != 1 ||
			   fwrite(++ p_raster, sizeof(unsigned char), 1, p_fw) != 1) {
				fclose(p_fw);
				return false;
			}
		}
		// write scanline

		static const unsigned char p_dummy[3] = {0};
		if(n_scanline_padd && fwrite(p_dummy, n_scanline_padd, 1, p_fw) != 1) {
			fclose(p_fw);
			return false;
		}
		// write padding
	}
	// write bitmap

	fclose(p_fw);

	return true;
}

int WINAPI WinMain(HINSTANCE _h_instance, HINSTANCE h_prev_instance,
	PSTR szCmdline, int nCmdShow)
{
	if(!Init_OpenGL()) {
		MessageBox(0, "OpenGL initialization failed", "Error", MB_OK);
		return -1;
	}
	// init openGL

	TAVI_Stream *p_video = p_AVI_Load("jerabnici.avi", true);
	if(p_video == NULL) {
		MessageBox(0, "error loading \'jerabnici.avi\'", "Error", MB_OK);
		Shutdown_OpenGL();
		return -1;
	}
	// open avi

	ImageStruct image_a, image_b, image_c;
	image_a.Raster = NULL; // initialize!
	image_a.XSize = p_video->n_width;
	image_a.YSize = p_video->n_height;
	image_a.XOffset = 4;
	image_a.YOffset = 4 * p_video->n_width;
	image_a.PixelType = ImageRGBLinear;
	image_c = image_b = image_a;
	if((image_a.Raster = (unsigned char*)malloc(4 * p_video->n_width * p_video->n_height)) == NULL ||
	   (image_b.Raster = (unsigned char*)malloc(4 * p_video->n_width * p_video->n_height)) == NULL ||
	   (image_c.Raster = (unsigned char*)malloc(4 * p_video->n_width * p_video->n_height)) == NULL) {
		if(image_a.Raster != NULL)
			free(image_a.Raster);
		if(image_b.Raster != NULL)
			free(image_b.Raster);
		if(image_c.Raster != NULL)
			free(image_c.Raster);
		Shutdown_OpenGL();
		AVI_Free(p_video);
		MessageBox(0, "not enough memory for test images", "Error", MB_OK);
		return -1;
	}
	// alloc two images

	AVI_GetFrame_t(p_video, 10, image_a.Raster, 4, 8);
	AVI_GetFrame_t(p_video, 20, image_b.Raster, 4, 8);
	// get two images

	Save_RGBLinearImageStruct("image_a.bmp", &image_a);
	Save_RGBLinearImageStruct("image_b.bmp", &image_b);

	AVI_Free(p_video);
	// close avi

	bool b_errors = false;

	if(glGetError() != GL_NO_ERROR) {
		MessageBox(0, "GL error after init :(", "Error", MB_OK);
		b_errors = true;
	}
	if(b_ErrorLevel()) {
		MessageBox(0, "digilib-ext error after init :(", "Error", MB_OK);
		b_errors = true;
	}

	ImageImageAdd(&image_c, &image_a, &image_b);
	// try to add images (have to alloc buffers & textures)

	if(glGetError() != GL_NO_ERROR) {
		MessageBox(0, "GL error after first run :(", "Error", MB_OK);
		b_errors = true;
	}
	if(b_ErrorLevel()) {
		MessageBox(0, "digilib-ext error after first run :(", "Error", MB_OK);
		b_errors = true;
	}

	ImageImageAdd(&image_c, &image_a, &image_b);
	// try to add images (should be much smoother now)

	if(glGetError() != GL_NO_ERROR) {
		MessageBox(0, "GL error after second run :(", "Error", MB_OK);
		b_errors = true;
	}
	if(b_ErrorLevel()) {
		MessageBox(0, "digilib-ext error after second run :(", "Error", MB_OK);
		b_errors = true;
	}

	ImageStruct *image_d = p_Create_GL_Image(image_a.XSize, image_a.YSize, ImageRGBLinear);
	if(image_d != NULL) {
		CTimer my_timer;
		my_timer.ResetTimer();
		int n_frame_counter = 0;
		float f_last_time;
		while((f_last_time = my_timer.f_Time()) < 1) {
			ImageImageMul(image_d, &image_a, &image_b);
			Disable_AutoDownloads();
			++ n_frame_counter;
		}
		Enable_AutoDownloads();
		// try to multiply images (should be much smoother now)

		char p_s_speed_msg[256];
		sprintf(p_s_speed_msg, "rendering at %f frames / sec\n", n_frame_counter / f_last_time);
		MessageBox(0, p_s_speed_msg, "DigiLib-Ext", MB_OK);
		// print some speed stats

		if(glGetError() != GL_NO_ERROR) {
			MessageBox(0, "GL error after third run :(", "Error", MB_OK);
			b_errors = true;
		}
		if(b_ErrorLevel()) {
			MessageBox(0, "digilib-ext error after third run :(", "Error", MB_OK);
			b_errors = true;
		}

		Save_RGBLinearImageStruct("image_a_mul_b.bmp", image_d);

		ImageToGreyscale(image_d, &image_a);
		Save_RGBLinearImageStruct("image_a_grey_shader.bmp", image_d);

		if(glGetError() != GL_NO_ERROR) {
			MessageBox(0, "GL error after shader run :(", "Error", MB_OK);
			b_errors = true;
		}
		if(b_ErrorLevel()) {
			MessageBox(0, "digilib-ext error after shader run :(", "Error", MB_OK);
			b_errors = true;
		}

		ImageImageAdd(image_d, &image_a, &image_b);
		Save_RGBLinearImageStruct("image_a_post_shader_add_b.bmp", image_d);

		if(glGetError() != GL_NO_ERROR) {
			MessageBox(0, "GL error after post-shader run :(", "Error", MB_OK);
			b_errors = true;
		}
		if(b_ErrorLevel()) {
			MessageBox(0, "digilib-ext error after post-shader run :(", "Error", MB_OK);
			b_errors = true;
		}

		Destroy_GL_Image(image_d);
	} else
		MessageBox(0, "unable to proceed with the third run :(", "Error", MB_OK);

	free(image_a.Raster);
	free(image_b.Raster);
	free(image_c.Raster);
	// free images

	if(glGetError() != GL_NO_ERROR) {
		MessageBox(0, "GL error(s) occured", "Error", MB_OK);
		b_errors = true;
	}

	if(!b_errors)
		MessageBox(0, "al tests ok", "DigiLib-Ext", MB_OK);

	if(!Shutdown_OpenGL()) {
		MessageBox(0, "OpenGL shutdown failed", "Error", MB_OK);
		return -1;
	}

	return 0;
}

/*
 *		-end-of-file-
 */
