/*
								+----------------------------------+
								|                                  |
								|   *** Renderbuffer manager ***   |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|            Manager.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __GL_RENDERBUFFER_MANAGER_INCLUDED
#define __GL_RENDERBUFFER_MANAGER_INCLUDED

#include "Manager.h"

/*
 *	class CGLNamedRenderBuffer_FBO
 *		- extension of render-buffer with automatically assigned id
 */
class CGLNamedRenderBuffer_FBO : public CGLRenderBuffer_FBO {
protected:
	static int m_n_id_space;
	int m_n_id;

public:
	inline int n_Id() const { return m_n_id; }
	inline GLenum n_Pixel_Format() const { return m_n_color_internal_format; }

	inline CGLNamedRenderBuffer_FBO(int n_width, int n_height,
		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
		:CGLRenderBuffer_FBO(n_width, n_height,
		 n_color_buffer_num, b_color_texture_target, n_color_internal_format,
		 b_depth_buffer, b_depth_texture_target, n_depth_internal_format,
		 b_stencil_buffer, b_stencil_texture_target, n_stencil_internal_format)
	{
		m_n_id = ++ m_n_id_space;
	}
};

//int CGLNamedRenderBuffer_FBO::m_n_id_space = 0;
// needs to be declared once

/*
 *	class CGetFBOCost
 *		- cost evaluator for render-buffer manager purging purposes
 *		- currently returns size of buffer surface in pixels
 */
class CGetFBOCost {
public:
	inline int n_Cost(const CGLNamedRenderBuffer_FBO *p_fbo) const
	{
		return p_fbo->n_Width() * p_fbo->n_Height(); // todo - calculate with pixel size
	}
};

/*
 *	class CFindRB_Format
 *		- finds render-buffer by 2d-texture it's necessary to be rendered to
 */
class CFindRB_Format {
protected:
	const CGLTexture_2D *m_p_format;

public:
	inline CFindRB_Format(const CGLTexture_2D *p_format)
		:m_p_format(p_format)
	{}

	inline bool operator ()(const CGLTexture_2D *p_fbo) const
	{
		return p_fbo->n_Width() == m_p_format->n_Width() &&
			p_fbo->n_Height() == m_p_format->n_Height() &&
			p_fbo->n_Pixel_Format() == m_p_format->n_Pixel_Format();
	}
};

/*
 *	class CRenderBufferFactory
 *		- render-buffer factory
 */
class CRenderBufferFactory {
public:
	inline CGLNamedRenderBuffer_FBO *Load(const CGLTexture_2D *p_format)
	{
		return new CGLNamedRenderBuffer_FBO(p_format->n_Width(), p_format->n_Height(),
			1, true, p_format->n_Pixel_Format(), false, false, 0, false, false, 0);
	}

	static inline bool b_Valid_Object(const CGLNamedRenderBuffer_FBO *p_object)
	{
		return p_object != 0;
	}
};

/*
 *	class CGLRenderBuffer_Manager
 *		- OpenGL render buffer manager class
 */
typedef CResourceManager<CGLNamedRenderBuffer_FBO, CGetFBOCost, int, CFindIntId<CGLNamedRenderBuffer_FBO>,
	const CGLTexture_2D*, CFindRB_Format, CRenderBufferFactory> CGLRenderBuffer_Manager;

#endif //__GL_RENDERBUFFER_MANAGER_INCLUDED
