/*
								+----------------------------------+
								|                                  |
								| *** DigiLib OpenGL extension *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|          ShadersSrc.cpp          |
								|                                  |
								+----------------------------------+
*/

#if defined(_MSC_VER) && !defined(__MWERKS__)
#pragma warning(disable:4786)
#endif
// disable warning C4786 "identifier was truncated to '255' characters in the debug information"

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <gl/gl.h>
#include <gl/glut.h>
#include <gl/glext.h>
#include <algorithm>
#include <stdlib.h>
#include <string.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLDrv.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/VertexBufferObject.h"
#include "Shader_Manager.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

class CAdHocShaderInfo : public TShaderInfo {
	TShaderInfo::TShader m_t_high;

public:
	CAdHocShaderInfo(const char *p_s_shader_name, bool b_fragment_shader,
		const char *p_s_source_code, const char *p_s_sampler_names = 0)
	{
		p_s_name = (char*)p_s_shader_name;
		b_fragment = b_fragment_shader; // vertex or fragment?
		b_shader_target = true; // "shader", not "program"
		p_low = 0; // no program
		p_high = &m_t_high; // my shader

		m_t_high.p_s_source_code = (char*)p_s_source_code;
		m_t_high.n_sampler_num = 0;
		if(p_s_sampler_names) {
			for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
			   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0)
				++ m_t_high.n_sampler_num;
			// count samplers

			m_t_high.p_sampler = new TShaderInfo::TShader::TSamplerInfo[m_t_high.n_sampler_num];
			_ASSERTE(m_t_high.p_sampler);
			// alloc samplers

			m_t_high.n_sampler_num = 0;
			for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
			   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0) {
				char *p_s_str;
				if(strchr(p_s_sampler, '|')) {
					int n_size = strchr(p_s_sampler, '|') - p_s_sampler;
					if(p_s_str = new char[n_size + 1]) {
						strncpy(p_s_str, p_s_sampler, n_size);
						p_s_str[n_size] = 0;
					}
				} else
					p_s_str = p_s_Alloc(p_s_sampler);
				_ASSERTE(p_s_str);
				// get a single sampler name

				m_t_high.p_sampler[m_t_high.n_sampler_num].p_s_name = p_s_str;
				m_t_high.p_sampler[m_t_high.n_sampler_num].n_resource_index = m_t_high.n_sampler_num;
				// assign samplers

				++ m_t_high.n_sampler_num;
			}
			// assign sampler names and indices
		} else
			m_t_high.p_sampler = 0;
	}

	~CAdHocShaderInfo()
	{
		m_t_high.p_s_source_code = 0;
		// don't want to be deleted by TShaderInfo::TShader destructor as it's static data

		p_s_name = 0;
		p_high = 0;
		// don't want them to be deleted by TShaderInfo destructor as they are static data
	}
};

CAdHocShaderInfo my_test_shader("my_test_shader", true, "void main()\n{\ngl_FragColor = vec4(1, 0, 0, 1);\n// test red output\n}\n");
// dummy test shader (const red output)

CAdHocShaderInfo my_test_shader2("my_test_shader2", true,
	"uniform sampler2D operand1, operand2;\n"
	"void main()\n"
	"{\n"
	"    gl_FragColor = texture2D(operand1, gl_TexCoord[0].xy) - texture2D(operand2, gl_TexCoord[0].xy);\n"
	"    // test red output\n"
	"}\n", "operand1|operand2");
// test shader 2 (subtracts two operands)

CAdHocShaderInfo greyscale_shader("greyscale_shader", true,
	"uniform sampler2D operand1;\n"
	"void main()\n"
	"{\n"
	"    float f_grey = dot(texture2D(operand1, gl_TexCoord[0].xy),\n"
	"        vec4(.299, .587, .144, 0));\n"
	"    gl_FragColor = vec4(f_grey, f_grey, f_grey, 1);\n"
	"}\n");
// greyscale shader

CAdHocShaderInfo color_matrix_shader("color-matrix_shader", true,
	"uniform sampler2D operand1;\n"
	"uniform mat4 matrix;\n"
	"void main()\n"
	"{\n"
	"    gl_FragColor = texture2D(operand1, gl_TexCoord[0].xy) * matrix;\n"
	"}\n");
// color matrix shader

/*
 *	bool Register_Shaders()
 *		- registers shader source codes with shader names to CShaderFactory
 *		- returns true on success, false on failure (not enough memory to hold shader list)
 */
bool Register_Shaders()
{
	/*if(!CShaderFactory::Register_ShaderSource("|my_test_shader", 0, &my_test_shader))
		return false;

	if(!CShaderFactory::Register_ShaderSource("|my_test_shader2", 0, &my_test_shader2))
		return false;*/ // test shaders only

	if(!CShaderFactory::Register_ShaderSource("|greyscale_shader", 0, &greyscale_shader))
		return false;

	if(!CShaderFactory::Register_ShaderSource("|color-matrix_shader", 0, &color_matrix_shader))
		return false;

	return true;
}

/*
 *		-end-of-file-
 */
